from __future__ import annotations

from pyjlyric.model import LyricPage, WithUrlText


class UtatenLyricPage(LyricPage):
    artist: str | WithUrlText
    composer: str | WithUrlText
    lyricist: str | WithUrlText
    arranger: None | str | WithUrlText = None



"""<https://linkco.re>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.util import (
    get_captured_value,
    get_source,
    parse_obj_as_url,
    parse_text_with_optional_link,
    select_one_tag,
)

from .model import LinkcoreLyricPage

_LINKCORE_PATTERN = r"^https://linkco\.re/(?P<albumid>[a-zA-Z0-9]+)/songs/(?P<songid>[0-9]+)/lyrics$"


class LinkcoreLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class LinkcoreLyricPageParser(BaseLyricPageParser):
    """https://linkco.re/<albumid>/songs/<songid>/lyrics"""

    _test = "https://linkco.re/0aQUYTE5/songs/1825986/lyrics"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_LINKCORE_PATTERN)
        m = re.match(pattern, url)
        albumid = get_captured_value(m, "albumid")
        songid = get_captured_value(m, "songid")

        return albumid is not None and songid is not None

    @staticmethod
    def parse(url: str) -> LinkcoreLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_LINKCORE_PATTERN)
        m = re.match(pattern, url)
        albumid = get_captured_value(m, "albumid")
        if albumid is None:
            raise LinkcoreLyricPageParserError from ValueError
        songid = get_captured_value(m, "songid")
        if songid is None:
            raise LinkcoreLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise LinkcoreLyricPageParserError from ConnectionError

        title_h2 = select_one_tag(bs, "div.header_text > h2")

        artist_h3 = select_one_tag(bs, "div.header_text > h3")
        artist_a = artist_h3.a
        artist_link = (
            None
            if artist_a is None or artist_a.href is None
            else parse_obj_as_url(
                artist_a.href.text,
                base=url,
            )
        )

        lyricist_p = select_one_tag(bs, "div.lyric_credit > ul > li:nth-child(1) > p")
        composer_p = select_one_tag(bs, "div.lyric_credit > ul > li:nth-child(2) > p")

        lyric_sections: list[list[str]] = [[]]
        for section in select_one_tag(bs, "div.lyric_text").find_all("p"):
            if section == "":
                lyric_sections.append([])
                continue
            lyric_sections[-1].append(section.text)

        return LinkcoreLyricPage(
            title=title_h2.text,
            page_url=parse_obj_as_url(url),
            pageid=f"{albumid}-{songid}",
            artist=(
                artist_h3.text.split(", ") if artist_h3 else parse_text_with_optional_link(artist_h3.text, artist_link)
            ),
            composer=composer_p.text.split(", "),
            lyricist=lyricist_p.text.split(", "),
            arranger=None,
            lyric_sections=lyric_sections,
        )



"""<https://lyric.evesta.com>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import convert_lines_into_sections, get_captured_value, get_source, parse_obj_as_url, select_one_tag

from .model import GakkiLyricPage

_GAKKI_PATTERN = r"^https://gakufu.gakki.me/m/data/(?P<pageid>[A-Z]+\d+)\.html$"


class GakkiLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class GakkiLyricPageParser(BaseLyricPageParser):
    """https://gakufu.gakki.me/m/data/<pageid>.html"""

    _test = "https://gakufu.gakki.me/m/data/RQ17577.html"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_GAKKI_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> GakkiLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_GAKKI_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise GakkiLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise GakkiLyricPageParserError from ConnectionError
        m = re.match(r"^作詞\s(?P<lyricist>.*)\s作曲\s(?P<composer>.*)$", select_one_tag(bs, "div.info > p.data").text)
        if m is None:
            raise GakkiLyricPageParserError from ValueError

        lyricist, composer = str(m.group("lyricist")), str(m.group("composer"))

        artist = select_one_tag(bs, "#topics > ol > li:nth-child(2) > a")

        lyric_chars = []
        for div in bs.select("div#chord_area > div"):
            if div.get("class") is None:
                lyric_chars.append("\n")
            else:
                lyric_chars.extend(d.text for d in div.select("div.cd_pic.blue > div") if not d.text.strip())

        lyric_lines = [i.replace("\n", "") for i in "\n".join(lyric_chars).strip().split("\n\n\n")]

        return GakkiLyricPage(
            title=select_one_tag(bs, "div.info > h2.tit > span").text[1:-1],
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=WithUrlText(text=artist.text.strip(), link=parse_obj_as_url(str(artist.get("href")), base=url)),
            composer=composer,
            lyricist=lyricist,
            arranger=None,
            lyric_sections=convert_lines_into_sections(lyric_lines),
        )



"""Utility functions."""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup, Tag
from pydantic import HttpUrl, TypeAdapter

from pyjlyric.model import WithUrlText

if TYPE_CHECKING:
    from re import Match

_UA = """
Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko)
Chrome/110.0.0.0 Safari/537.36
""".strip().replace(
    "\n",
    " ",
)

_REQUESTS_TIMEOUT = 10
_REQUESTS_HEADERS = {
    "User-Agent": _UA,
}


def get_captured_value(m: Match[str] | None, group_name: str) -> str | None:
    """Get the captured value with the group name from the match object.

    Parameters
    ----------
    m : Match | None
        matched result
    group_name : str | int
        group name

    Returns
    -------
    str | None
        return matched string if exists, otherwise None
    """
    if m is None or group_name not in (d := m.groupdict()):
        return None

    return str(d[group_name])


def get_source(
    url: str,
    *,
    data: dict[str, str | int] | None = None,
    method: Literal["get", "post"] | None = "get",
    headers: dict[str, str] | None = None,
    parser: str | None = "lxml",
) -> BeautifulSoup | None:
    """Get the source as a BeautifulSoup object.

    Parameters
    ----------
    url : str
        Web Page URL

    Returns
    -------
    BeautifulSoup | None
        parsed source data of web page
    """
    headers = _REQUESTS_HEADERS if headers is None else dict(_REQUESTS_HEADERS, **headers)

    if method is None or method == "get":
        res = requests.get(url, data=data, timeout=_REQUESTS_TIMEOUT, headers=headers)
    elif method == "post":
        res = requests.post(url, data=data, timeout=_REQUESTS_TIMEOUT, headers=headers)
    else:
        raise ValueError(method)

    if not res.ok:
        return None
    return BeautifulSoup(
        markup=res.content if res.content != b"" else res.text,
        features="lxml" if parser is None else parser,
    )


def select_one_tag(bs: BeautifulSoup | Tag, selector: str) -> Tag:
    """WIP."""
    res = bs.select_one(selector)

    if res is None:
        raise ValueError(res)

    return res


def parse_obj_as_url(url: str, *, base: str | None = None) -> HttpUrl:
    """WIP."""
    if base is not None:
        url = urljoin(base, url)
    return TypeAdapter(HttpUrl).validate_python(url)


def parse_text_with_optional_link(text: str, link: HttpUrl | None) -> WithUrlText | str:
    if link is None:
        return text
    return WithUrlText(text=text, link=link)


def convert_lines_into_sections(lines: list[str]) -> list[list[str]]:
    lyric_sections = []
    now: list[str] = []
    for line in [*lines, ""]:
        if not line:
            if len(now) > 0:
                lyric_sections.append(now)
            now = []
        else:
            now.append(line)
    return lyric_sections



"""<https://l-lyric.net>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.util import (
    get_captured_value,
    get_source,
    parse_obj_as_url,
    parse_text_with_optional_link,
    select_one_tag,
)

from .model import JlyricLyricPage

_JLYRIC_PATTERN = r"^https://j-lyric\.net/artist/(?P<artistid>[a-z0-9]+)/(?P<pageid>[a-z0-9]+)\.html$"


class JlyricLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class JlyricLyricPageParser(BaseLyricPageParser):
    """https://j-lyric.net/artist/<artistid>/<pageid>.html"""

    _test = "https://j-lyric.net/artist/a05d28d/l05c028.html"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_JLYRIC_PATTERN)
        m = re.match(pattern, url)
        artistid = get_captured_value(m, "artistid")
        pageid = get_captured_value(m, "pageid")

        return artistid is not None and pageid is not None

    @staticmethod
    def parse(url: str) -> JlyricLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_JLYRIC_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise JlyricLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise JlyricLyricPageParserError from ConnectionError

        title_div = select_one_tag(bs, "#mnb > div.cap")
        title_m = re.match("^「(.*)」歌詞$", title_div.text)
        if title_m is None or len(title_m.groups()) != 1:
            raise JlyricLyricPageParserError from ValueError

        artist = select_one_tag(bs, "#mnb > div.lbdy > p:nth-child(1) > a")
        artist_link = (
            None
            if artist.href is None
            else parse_obj_as_url(
                artist.href.text,
                base=url,
            )
        )

        credit_p = select_one_tag(bs, "#mnb > div.lbdy > p:nth-child(2)")
        credit_text = credit_p.text

        lyricist_m = re.search(r"作詞\uff1a([^\u3000\s]+)", credit_text)
        if lyricist_m is None or len(lyricist_m.groups()) != 1:
            raise JlyricLyricPageParserError from ValueError

        composer_m = re.search(r"作曲\uff1a(.+)$", credit_text)
        if composer_m is None or len(composer_m.groups()) != 1:
            raise JlyricLyricPageParserError from ValueError

        lyric = select_one_tag(bs, "#Lyric")
        for br in lyric.select("br"):
            br.replace_with("\n")

        return JlyricLyricPage(
            title=title_m.groups()[0],
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=parse_text_with_optional_link(artist.text, artist_link),
            composer=composer_m.groups()[0],
            lyricist=lyricist_m.groups()[0],
            arranger=None,
            lyric_sections=[section.strip().split("\n") for section in lyric.text.replace("\u3000", " ").split("\n\n")],
        )



"""<https://lyric.evesta.com>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.util import (
    get_captured_value,
    get_source,
    parse_obj_as_url,
    parse_text_with_optional_link,
    select_one_tag,
)

from .model import EvestaLyricPage

_EVESTA_PATTERN = r"^https://lyric\.evesta\.jp/(?P<pageid>[a-z0-9]+)\.html$"


class EvestaLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class EvestaLyricPageParser(BaseLyricPageParser):
    """https://lyric.evesta.jp/<pageid>.html"""

    _test = "https://lyric.evesta.jp/l791684.html"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_EVESTA_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> EvestaLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_EVESTA_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise EvestaLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise EvestaLyricPageParserError from ConnectionError

        header_div = select_one_tag(bs, "div#lyrictitle")

        artist = select_one_tag(header_div, "#lyrictitle > p:nth-child(2) > a")
        artist_link = (
            None
            if artist is None or artist.href is None
            else parse_obj_as_url(
                artist.href.text,
                base=url,
            )
        )

        lyricist = select_one_tag(header_div, "#lyrictitle > p:nth-child(3)")

        composer = select_one_tag(header_div, "#lyrictitle > p:nth-child(4)")

        lyric = select_one_tag(bs, "div#lyricbody")
        for br in lyric.select("br"):
            br.replace_with("")

        return EvestaLyricPage(
            title=re.sub(r"\s歌詞$", "", select_one_tag(header_div, "h1").text),
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=parse_text_with_optional_link(artist.text, artist_link),
            composer=composer.text,
            lyricist=lyricist.text,
            arranger=None,
            lyric_sections=[section.strip().split("\r\n") for section in lyric.text.strip().split("\n\r\n")],
        )



"""<https://lyric.evesta.com>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import get_captured_value, get_source, parse_obj_as_url, select_one_tag

from .model import HoickLyricData, HoickLyricPage

_HOICK_PATTERN = r"^https://hoick\.jp/mdb/detail/(?P<pageid>\d+)/?.*$"


class HoickLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class HoickLyricPageParser(BaseLyricPageParser):
    """https://hoick.jp/mdb/detail/<pageid>"""

    _test = "https://hoick.jp/mdb/detail/5144/%e3%82%b1%e3%83%ad%e3%83%9d%e3%83%b3%e4%bd%93%e6%93%8d"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_HOICK_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> HoickLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_HOICK_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise HoickLyricPageParserError from ValueError

        bs = get_source(url, parser="html.parser")
        if bs is None:
            raise HoickLyricPageParserError from ConnectionError

        header_div = select_one_tag(bs, "div.m_bottom")

        lyricists = [
            WithUrlText(text=a.text, link=parse_obj_as_url(a.href.text, base=url))
            for a in select_one_tag(header_div, "h2.lylic").select("a")
            if a.href is not None
        ]

        composers = [
            WithUrlText(text=a.text, link=parse_obj_as_url(a.href.text, base=url))
            for a in select_one_tag(header_div, "h2.music").select("a")
            if a.href is not None
        ]

        bs_lyric = get_source(f"https://hoick.jp/data.php?id={pageid}&type=4")
        if bs_lyric is None:
            raise HoickLyricPageParserError from ConnectionError
        lyric_lines = HoickLyricData.model_validate_json(bs_lyric.text)

        return HoickLyricPage(
            title=select_one_tag(bs, "h1.detail").text,
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=None,
            composer=composers,
            lyricist=lyricists,
            arranger=None,
            lyric_sections=[section.strip().split("\r") for section in "".join(lyric_lines.root[:-1]).split("\r\r")],
        )



"""<https://www.joysound.com>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import get_captured_value, get_source, parse_obj_as_url

from .model import JoysoundLyricData, JoysoundLyricPage

_JOYSOUND_PATTERN = r"^https://www.joysound.com/web/search/song/(?P<pageid>\d+)$"


class JoysoundLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class JoysoundLyricPageParser(BaseLyricPageParser):
    """https://www.joysound.com/web/search/song/<pageid>"""

    _test = "https://www.joysound.com/web/search/song/112942"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_JOYSOUND_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> JoysoundLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_JOYSOUND_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise JoysoundLyricPageParserError from ValueError

        bs = get_source(
            "https://mspxy.joysound.com/Common/Lyric",
            method="post",
            data={
                "kind": "naviGroupId",
                "selSongNo": pageid,
                "interactionFlg": "0",
                "apiVer": "1.0",
            },
            headers={
                "x-jsp-app-name": "0000800",
            },
        )
        if bs is None:
            raise JoysoundLyricPageParserError from ConnectionError

        track_data = JoysoundLyricData.model_validate_json(bs.text)

        return JoysoundLyricPage(
            title=track_data.song_name,
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=WithUrlText(
                text=track_data.artist_name,
                link=parse_obj_as_url(
                    f"/web/search/artist/{track_data.artist_id}",
                    base=url,
                ),
            ),
            composer=track_data.composer,
            lyricist=track_data.lyricist,
            arranger=None,
            lyric_sections=[line.split("\n") for line in track_data.lyric_list[0].lyric.strip().split("\n\n")],
        )



from __future__ import annotations

from pyjlyric.model import LyricPage, WithUrlText


class LinkcoreLyricPage(LyricPage):
    artist: list[str] | WithUrlText | str
    composer: list[str]
    lyricist: list[str]



"""Abstract classes."""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing_extensions import Self

    from .model import LyricPage


class BaseLyricPageParserError(Exception): ...


class BaseLyricPageParser(ABC):
    _test: str

    def __init__(self: Self, url: str) -> None:
        if self.is_valid_url(url):
            self.data = self.parse(url)
            self.url = url

    @staticmethod
    @abstractmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def parse(url: str) -> LyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        raise NotImplementedError



"""<https://utaten.com>"""

from __future__ import annotations

import re

from bs4 import Tag

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.util import (
    get_captured_value,
    get_source,
    parse_obj_as_url,
    parse_text_with_optional_link,
    select_one_tag,
)

from .model import UtatenLyricPage

_UTATEN_PATTERN = r"^https://utaten\.com/lyric/(?P<pageid>[a-z]{2}\d+)/?$"


class UtatenLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class UtatenLyricPageParser(BaseLyricPageParser):
    """https://utaten.com/lyric/<pageid>"""

    _test = "https://utaten.com/lyric/ma21111253/"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_UTATEN_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> UtatenLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_UTATEN_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise UtatenLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise UtatenLyricPageParserError from ConnectionError

        title_h2 = select_one_tag(bs, "h2.newLyricTitle__main")
        for span in title_h2.select("span.newLyricTitle_afterTxt"):
            span.replace_with("")

        artist = select_one_tag(bs, "dt.newLyricWork__name > h3 > a")
        artist_link = (
            None
            if artist.href is None
            else parse_obj_as_url(
                artist.href.text,
                base=url,
            )
        )

        composer_text = bs.find(string="作曲")
        if composer_text is None:
            raise UtatenLyricPageParserError from TypeError
        composer_dt = composer_text.parent
        if not isinstance(composer_dt, Tag):
            raise UtatenLyricPageParserError from TypeError

        composer_dd = composer_dt.findNext("dd")
        if not isinstance(composer_dd, Tag):
            raise UtatenLyricPageParserError from TypeError

        composer_link = (
            None
            if composer_dd.href is None
            else parse_obj_as_url(
                composer_dd.href.text,
                base=url,
            )
        )

        lyricist_text = bs.find(string="作詞")
        if lyricist_text is None:
            raise UtatenLyricPageParserError from TypeError
        lyricist_dt = lyricist_text.parent
        if not isinstance(lyricist_dt, Tag):
            raise UtatenLyricPageParserError from TypeError

        lyricist_dd = lyricist_dt.findNext("dd")
        if not isinstance(lyricist_dd, Tag):
            raise UtatenLyricPageParserError from TypeError

        lyricist_link = (
            None
            if lyricist_dd.href is None
            else parse_obj_as_url(
                lyricist_dd.href.text,
                base=url,
            )
        )

        arranger_text = bs.find(string="編曲")
        arranger_dt = None if arranger_text is None else arranger_text.parent
        if not isinstance(arranger_dt, Tag):
            arranger_dt = None

        arranger_dd = None if arranger_dt is None else lyricist_dt.findNext("dd")
        if not isinstance(arranger_dd, Tag):
            arranger_dd = None

        arranger_link = (
            None
            if arranger_dd is None or arranger_dd.href is None
            else parse_obj_as_url(
                arranger_dd.href.text,
                base=url,
            )
        )

        lyric = select_one_tag(bs, "div.lyricBody > div > div.hiragana")
        for span in lyric.select("span.rt"):
            span.replace_with("")

        return UtatenLyricPage(
            title=title_h2.text.strip(),
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=parse_text_with_optional_link(artist.text.strip(), artist_link),
            composer=parse_text_with_optional_link(composer_dd.text.strip(), composer_link),
            lyricist=parse_text_with_optional_link(lyricist_dd.text.strip(), lyricist_link),
            arranger=None if arranger_dd is None else parse_text_with_optional_link(arranger_dd.text, arranger_link),
            lyric_sections=[section.strip().split("\n") for section in lyric.text.replace("\u3000", " ").split("\n\n")],
        )



from __future__ import annotations

from typing import Literal

from pydantic import BaseModel, Field

from pyjlyric.model import LyricPage, WithUrlText


class MusicbookLyricPage(LyricPage):
    artist: WithUrlText
    composer: str | None = None
    lyricist: str | None = None
    arranger: None


class MusicbookLyricData(BaseModel):
    title: str | None = None
    writer: str | None = None
    composer: str | None = None
    lyrics: list[str]
    error_code: Literal["apiCallError"] | None = Field(None, alias="errorCode")



from __future__ import annotations

from pyjlyric.model import LyricPage, WithUrlText


class GakkiLyricPage(LyricPage):
    artist: WithUrlText
    composer: str
    lyricist: str
    arranger: None



from __future__ import annotations

from typing import TYPE_CHECKING

from pydantic import RootModel, field_validator
from typing_extensions import Self

from pyjlyric.model import LyricPage, WithUrlText

if TYPE_CHECKING:
    from collections.abc import Iterator


class HoickLyricPage(LyricPage):
    artist: None
    composer: list[WithUrlText]
    lyricist: list[WithUrlText]
    arranger: None


class HoickLyricData(RootModel[list[str]]):
    def __iter__(self: Self) -> Iterator[str]:  # type: ignore[override]
        return iter(self.root)

    @field_validator("root")
    @classmethod
    def validate(cls: type[Self], v: list[str]) -> list[str]:  # type: ignore[override]
        return [i for i in v if isinstance(i, str)]



from __future__ import annotations

import textwrap
from os import linesep
from typing import TYPE_CHECKING

from pyjlyric.model import WithUrlText

from .animap.parser import AnimapLyricPageParser
from .evesta.parser import EvestaLyricPageParser
from .gakki.parser import GakkiLyricPageParser
from .hoick.parser import HoickLyricPageParser
from .jlyric.parser import JlyricLyricPageParser
from .joysound.parser import JoysoundLyricPageParser
from .jtotal.parser import JtotalLyricPageParser
from .kashinavi.parser import KashinaviLyricPageParser
from .linkcore.parser import LinkcoreLyricPageParser
from .musicbook.parser import MusicbookLyricPageParser
from .petitlyrics.parser import PetitlyricsLyricPageParser
from .utamap.parser import UtamapLyricPageParser
from .utanet.parser import UtanetLyricPageParser
from .utaten.parser import UtatenLyricPageParser

if TYPE_CHECKING:
    from pyjlyric.model import LyricPage

    from .base import BaseLyricPageParser


Parsers: tuple[type[BaseLyricPageParser], ...] = (
    AnimapLyricPageParser,
    EvestaLyricPageParser,
    GakkiLyricPageParser,
    HoickLyricPageParser,
    KashinaviLyricPageParser,
    JoysoundLyricPageParser,
    JlyricLyricPageParser,
    JtotalLyricPageParser,
    LinkcoreLyricPageParser,
    MusicbookLyricPageParser,
    PetitlyricsLyricPageParser,
    UtamapLyricPageParser,
    UtanetLyricPageParser,
    UtatenLyricPageParser,
)


def parse(url: str) -> LyricPage | None:
    for parser in Parsers:
        if parser.is_valid_url(url):
            return parser.parse(url)
    return None


def get_lyric_text(url: str) -> str | None:
    """CLI main."""
    r = parse(url)

    if r is None:
        return None

    title = r.title
    artist = r.artist.text if isinstance(r.artist, WithUrlText) else r.artist
    artist = (
        ", ".join([a.text if isinstance(a, WithUrlText) else a for a in artist]) if isinstance(artist, list) else artist
    )

    lyricist = r.lyricist.text if isinstance(r.lyricist, WithUrlText) else r.lyricist
    lyricist = (
        ", ".join([ly.text if isinstance(ly, WithUrlText) else ly for ly in lyricist])
        if isinstance(lyricist, list)
        else lyricist
    )

    composer = r.composer.text if isinstance(r.composer, WithUrlText) else r.composer
    composer = (
        ", ".join([c.text if isinstance(c, WithUrlText) else c for c in composer])
        if isinstance(composer, list)
        else composer
    )

    lyric_sections = [linesep.join(paragraph) for paragraph in r.lyric_sections]

    return (
        textwrap.dedent(
            f"""
            ===
            Title:\t\t{title}
            Artist:\t\t{artist or "(No data)"}
            Lyric:\t\t{lyricist or "(No data)"}
            Composer:\t{composer or "(No data)"}
            ===
            """,
        )
        + (linesep * 2).join(lyric_sections).strip()
    ).strip()



"""<https://uta-net.com>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.util import (
    get_captured_value,
    get_source,
    parse_obj_as_url,
    parse_text_with_optional_link,
    select_one_tag,
)

from .model import UtanetLyricPage

_UTANET_PATTERN = r"^https://www.uta-net.com/song/(?P<pageid>\d+)/?$"


class UtanetLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class UtanetLyricPageParser(BaseLyricPageParser):
    """https://www.uta-net.com/song/<pageid>"""

    _test = "https://www.uta-net.com/song/155003/"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_UTANET_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> UtanetLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_UTANET_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise UtanetLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise UtanetLyricPageParserError from ConnectionError

        header_div = select_one_tag(bs, "div.song-infoboard")

        artist = select_one_tag(header_div, "a[itemprop='byArtist']")
        artist_link = (
            None
            if artist.href is None
            else parse_obj_as_url(
                artist.href.text,
                base=url,
            )
        )

        composer = select_one_tag(header_div, "a[itemprop='composer']")
        composer_link = (
            None
            if composer.href is None
            else parse_obj_as_url(
                composer.href.text,
                base=url,
            )
        )

        lyricist = select_one_tag(header_div, "a[itemprop='lyricist']")
        lyricist_link = (
            None
            if lyricist.href is None
            else parse_obj_as_url(
                lyricist.href.text,
                base=url,
            )
        )

        arranger = header_div.select_one("a[itemprop='arranger']")
        arranger_link = (
            None
            if arranger is None or arranger.href is None
            else parse_obj_as_url(
                arranger.href.text,
                base=url,
            )
        )

        lyric = select_one_tag(bs, "div#kashi_area")
        for br in lyric.select("br"):
            br.replace_with("\n")

        return UtanetLyricPage(
            title=select_one_tag(header_div, "h2").text,
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=parse_text_with_optional_link(artist.text, artist_link),
            composer=parse_text_with_optional_link(composer.text, composer_link),
            lyricist=parse_text_with_optional_link(lyricist.text, lyricist_link),
            arranger=None if arranger is None else parse_text_with_optional_link(arranger.text, arranger_link),
            lyric_sections=[section.strip().split("\n") for section in lyric.text.replace("\u3000", " ").split("\n\n")],
        )



from __future__ import annotations

from pydantic import BaseModel, Field, HttpUrl

from pyjlyric.model import LyricPage, WithUrlText


class _ByArtist(BaseModel):
    type_: str = Field(..., alias="@type")
    name_: str
    url: HttpUrl


class _RecordedAs(BaseModel):
    type_: str = Field(..., alias="@type")
    by_artist: _ByArtist = Field(..., alias="byArtist")


class _Lyricist(BaseModel):
    type_: str = Field(..., alias="@type")
    name: str


class _Composer(BaseModel):
    type_: str = Field(..., alias="@type")
    name: str


class _MainEntityOfPage(BaseModel):
    type_: str = Field(..., alias="@type")
    url: HttpUrl


class _Lyrics(BaseModel):
    type_: str = Field(..., alias="@type")
    main_entity_of_page: _MainEntityOfPage = Field(..., alias="mainEntityOfPage")
    text: str


class KashinaviLyricPage(LyricPage):
    artist: WithUrlText
    composer: str
    lyricist: str
    arranger: None



"""<https://music-book.jp>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import convert_lines_into_sections, get_captured_value, get_source, parse_obj_as_url, select_one_tag

from .model import MusicbookLyricData, MusicbookLyricPage

_MUSICBOOK_PATTERN = r"^https://music-book\.jp/music/Artist/(?P<artistid>\d+)/Music/(?P<pageid>[a-z0-9]+)$"


class MusicbookLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class MusicbookLyricPageParser(BaseLyricPageParser):
    """https://music-book.jp/music/Artist/<artistid>/Music/<pageid>"""

    _test = "https://music-book.jp/music/Artist/1538923/Music/aaahiqc4"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_MUSICBOOK_PATTERN)
        m = re.match(pattern, url)
        artistid = get_captured_value(m, "artistid")
        pageid = get_captured_value(m, "pageid")

        return artistid is not None and pageid is not None

    @staticmethod
    def parse(url: str) -> MusicbookLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_MUSICBOOK_PATTERN)
        m = re.match(pattern, url)
        artistid = get_captured_value(m, "artistid")
        if artistid is None:
            raise MusicbookLyricPageParserError from ValueError

        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise MusicbookLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise MusicbookLyricPageParserError from ConnectionError

        title = select_one_tag(bs, "h1").text

        artist = select_one_tag(bs, "a.text-with-icon.artist")
        artist_name = artist.text.strip()

        m = re.search(r"href=\"(?P<link>/music/Artist/\d+)\"", str(artist))
        artist_link = get_captured_value(m, "link")
        if artist_link is None:
            raise MusicbookLyricPageParserError from ValueError

        lyricist = select_one_tag(bs, "li:nth-child(1) > div > span").text.strip()
        composer = select_one_tag(bs, "li:nth-child(2) > div > span").text.strip()

        bs_lyric = get_source(f"https://music-book.jp/music/Lyrics/Track?artistName={artist_name}&trackTitle={title}")
        if bs_lyric is None:
            raise MusicbookLyricPageParserError from ConnectionError
        lyric_data = MusicbookLyricData.model_validate_json(bs_lyric.text)

        return MusicbookLyricPage(
            title=title,
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=WithUrlText(text=artist_name, link=parse_obj_as_url(artist_link, base=url)),
            composer=composer if composer != "-" and not composer else lyric_data.composer,
            lyricist=lyricist if lyricist != "-" and not lyricist else lyric_data.writer,
            arranger=None,
            lyric_sections=convert_lines_into_sections(lyric_data.lyrics),
        )



"""A module for CLI implementation."""

from __future__ import annotations

import shutil
import sys
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    Namespace,
    RawDescriptionHelpFormatter,
)

from pyjlyric.parse import Parsers, get_lyric_text

from . import __version__


class HelpFormatter(
    ArgumentDefaultsHelpFormatter,
    RawDescriptionHelpFormatter,
):
    """Custom formatter class."""


def parse_args() -> Namespace:
    """Parse given command line arguments.

    Returns
    -------
        Namespace: argparse.Namespace
    """
    usage = ["supported sites:"]
    usage += sorted(str(parser.__doc__).replace("<", ":").replace(">", "") for parser in Parsers)

    parser = ArgumentParser(
        prog="jrc",
        description="get lyric data by URL.",
        formatter_class=(
            lambda prog: HelpFormatter(
                prog,
                width=shutil.get_terminal_size(fallback=(120, 50)).columns,
                max_help_position=40,
            )
        ),
        epilog="\n  - ".join(usage),
    )
    parser.add_argument(
        "url",
    )
    parser.add_argument(
        "-V",
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    return parser.parse_args()


def main() -> int:
    """CLI main."""
    args = parse_args()
    url = str(args.url)
    r = get_lyric_text(url)

    if r is None:
        print(f"Error: Invalid link or unsupported site - <{url}>.", file=sys.stderr)  # noqa: T201
        return 1
    print(r)  # noqa: T201
    return 0


if __name__ == "__main__":
    sys.exit(main())



"""<https://music.j-total.net>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import get_captured_value, get_source, parse_obj_as_url, select_one_tag

from .model import JtotalLyricPage

_JTOTAL_PATTERN = r"^https://music\.j-total\.net/data/(?P<pageid>\d{3}[a-z]+/\d{3}[a-zA-Z_]+/\d{3})\.html$"


class JtotalLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class JtotalLyricPageParser(BaseLyricPageParser):
    """https://music.j-total.net/data/<pageid>.html"""

    _test = "https://music.j-total.net/data/004e/006_Elephant_Kashimashi/095.html"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_JTOTAL_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> JtotalLyricPage:
        """Parse the url page and return the result as LyricPage instance."""

        def is_valid_lyric(line: str) -> bool:
            line = line.strip()
            return line not in ("", "/") and not re.match("on.#", line) and not re.match("/.*/", line)

        pattern = re.compile(_JTOTAL_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise JtotalLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise JtotalLyricPageParserError from ConnectionError
        artist, lyricist, composer = select_one_tag(bs, "div.box2 > h2").text.split("/")
        artist_link = select_one_tag(bs, "body > div.wrapper > div.main > a").get("href")
        if artist_link is None:
            raise JtotalLyricPageParserError from ValueError

        lyric = select_one_tag(bs, "body > div.wrapper > div.main > p > tt")
        for a in lyric.select("a"):
            a.replace_with("")
        lyric_sections = [line.strip().split("\n") for line in lyric.text.split("\n\n")]
        lyric_sections = [
            [line.replace("\u3000", " ") for ly in lyric_section if is_valid_lyric(line := ly.strip())]
            for lyric_section in lyric_sections
        ]

        return JtotalLyricPage(
            title=select_one_tag(bs, "div.box2 > h1").text.strip(),
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=WithUrlText(text=artist[2:], link=parse_obj_as_url(str(artist_link))),
            composer=composer,
            lyricist=lyricist,
            arranger=None,
            lyric_sections=[s for s in lyric_sections if len(s) > 0],
        )



from __future__ import annotations

from pydantic import BaseModel, Field

from pyjlyric.model import LyricPage, WithUrlText


class JoysoundLyricPage(LyricPage):
    artist: WithUrlText
    composer: str | None = None
    lyricist: str | None = None
    arranger: None


class _LyricListItem(BaseModel):
    service_type: str = Field(..., alias="serviceType")
    service_publish_date: str = Field(..., alias="ServicePublishDate")
    status_code: str = Field(..., alias="statusCode")
    lyric: str


class _ItunesInfo(BaseModel):
    track_id: str = Field(..., alias="trackId")
    song_image_url: str = Field(..., alias="songImageUrl")
    item_url: str = Field(..., alias="itemUrl")
    preview_url: str = Field(..., alias="previewUrl")
    lyricsync_shift_time: str = Field(..., alias="lyricsyncShiftTime")


class _YoutubeInfo(BaseModel):
    movie_id: str = Field(..., alias="movieId")
    lyricsync_shift_time: str = Field(..., alias="lyricsyncShiftTime")
    thumbnail_url: str = Field(..., alias="thumbnailUrl")
    thumbnail_url_medium: str = Field(..., alias="thumbnailUrlMedium")
    thumbnail_url_large: str = Field(..., alias="thumbnailUrlLarge")
    disp_flg: str = Field(..., alias="dispFlg")


class _AmazonInfo(BaseModel):
    song_image_url: str = Field(..., alias="songImageUrl")
    song_image_url_small: str = Field(..., alias="songImageUrlSmall")
    song_image_url_large: str = Field(..., alias="songImageUrlLarge")
    item_name: str = Field(..., alias="itemName")
    item_url: str = Field(..., alias="itemUrl")


class _WikipediaInfo(BaseModel):
    item_url: str = Field(..., alias="itemUrl")
    official_url: str = Field(..., alias="officialUrl")
    twitter_account: str = Field(..., alias="twitterAccount")
    facebook_account: str = Field(..., alias="facebookAccount")


class _Field121centerInfo(BaseModel):
    song_image_url: str = Field(..., alias="songImageUrl")
    artist_image_url: str = Field(..., alias="artistImageUrl")


class _OutsideUrlInfo(BaseModel):
    itunes_info: _ItunesInfo = Field(..., alias="itunesInfo")
    youtube_info: _YoutubeInfo = Field(..., alias="youtubeInfo")
    amazon_info: _AmazonInfo = Field(..., alias="amazonInfo")
    wikipedia_info: _WikipediaInfo = Field(..., alias="wikipediaInfo")
    jsm_info: list[str] = Field(..., alias="jsmInfo")
    field_121center_info: _Field121centerInfo = Field(..., alias="121centerInfo")


class JoysoundLyricData(BaseModel):
    navi_group_id: str = Field(..., alias="naviGroupId")
    song_id: str = Field(..., alias="songId")
    sel_song_no: str = Field(..., alias="selSongNo")
    song_name: str = Field(..., alias="songName")
    song_name_ruby: str = Field(..., alias="songNameRuby")
    song_name_ruby_sort: str = Field(..., alias="songNameRubySort")
    song_name_alphabet: str = Field(..., alias="songNameAlphabet")
    song_name_alphabet_search_sort: str = Field(..., alias="songNameAlphabetSearchSort")
    song_name_japanese: str = Field(..., alias="songNameJapanese")
    song_name_japanese_ruby: str = Field(..., alias="songNameJapaneseRuby")
    song_name_japanese_sort: str = Field(..., alias="songNameJapaneseSort")
    song_name_foreign: str = Field(..., alias="songNameForeign")
    song_name_foreign_search: str = Field(..., alias="songNameForeignSearch")
    song_name_foreign_sort: str = Field(..., alias="songNameForeignSort")
    lyricist: str
    composer: str
    artist_id: str = Field(..., alias="artistId")
    group_artist_id: str = Field(..., alias="groupArtistId")
    artist_seq: str = Field(..., alias="artistSeq")
    artist_name: str = Field(..., alias="artistName")
    artist_name_ruby: str = Field(..., alias="artistNameRuby")
    artist_name_ruby_sort: str = Field(..., alias="artistNameRubySort")
    artist_name_alphabet_search_sort: str = Field(..., alias="artistNameAlphabetSearchSort")
    artist_name_foreign: str = Field(..., alias="artistNameForeign")
    artist_name_foreign_search: str = Field(..., alias="artistNameForeignSearch")
    artist_name_foreign_sort: str = Field(..., alias="artistNameForeignSort")
    artist_dv: str = Field(..., alias="artistDv")
    myartist_flg: str = Field(..., alias="myartistFlg")
    lyric_list: list[_LyricListItem] = Field(..., alias="lyricList")
    outside_url_info: _OutsideUrlInfo = Field(..., alias="outsideUrlInfo")



"""Pydantic models for lyric data."""

from __future__ import annotations

from pydantic import BaseModel, HttpUrl


class WithUrlText(BaseModel):
    """WIP."""

    link: HttpUrl
    text: str


_LyricSection = list[str]


class LyricPage(BaseModel):
    """WIP."""

    title: str
    page_url: HttpUrl
    pageid: str

    artist: str | WithUrlText | list[str] | list[WithUrlText] | None = None
    composer: str | WithUrlText | list[str] | list[WithUrlText] | None = None
    lyricist: str | WithUrlText | list[str] | list[WithUrlText] | None = None
    arranger: str | WithUrlText | list[str] | list[WithUrlText] | None = None

    lyric_sections: list[_LyricSection]



"""<http://www.animap.jp>"""

from __future__ import annotations

import re

from bs4 import BeautifulSoup
from requests import session

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import convert_lines_into_sections, get_captured_value, parse_obj_as_url, select_one_tag

from .model import PetitlyricsLyricData, PetitlyricsLyricPage

_PETITLYRICS_PATTERN = r"^https://petitlyrics\.com/lyrics/(?P<pageid>\d+)$"


class PetitlyricsLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class PetitlyricsLyricPageParser(BaseLyricPageParser):
    """https://petitlyrics.com/lyrics/<pageid>"""

    _test = "https://petitlyrics.com/lyrics/3287202"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_PETITLYRICS_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> PetitlyricsLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_PETITLYRICS_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise PetitlyricsLyricPageParserError from ValueError

        sess = session()
        res = sess.get(url)
        if not res.ok:
            raise PetitlyricsLyricPageParserError from ConnectionError
        bs = BeautifulSoup(res.text, "lxml")
        song_data_area = select_one_tag(bs, "td > div.pure-g > div.pure-u-1 > div:nth-child(1)")

        artist = select_one_tag(song_data_area, "p > a:nth-child(2)")
        artist_link = artist.get("href")
        if not isinstance(artist_link, str):
            raise PetitlyricsLyricPageParserError from ValueError

        m = re.search("(作詞\uff1a|Artist:)(?P<name>[^\u00a0]+)\u00a0", song_data_area.text)
        lyricist = None if m is None else get_captured_value(m, "name")

        m = re.search("(作曲\uff1a|Composer:)(?P<name>[^\u00a0]+)\u00a0", song_data_area.text)
        composer = None if m is None else get_captured_value(m, "name")

        res = sess.get("https://petitlyrics.com/lib/pl-lib.js")
        if not res.ok:
            raise PetitlyricsLyricPageParserError from ConnectionError
        m = re.search(r"'X-CSRF-Token', '(?P<csrf_token>[a-z0-9]{32})'", res.text)

        res_ajax = sess.post(
            "https://petitlyrics.com/com/get_lyrics.ajax",
            data={
                "lyrics_id": pageid,
            },
            headers={
                "X-CSRF-Token": get_captured_value(m, "csrf_token"),
                "X-Requested-With": "XMLHttpRequest",
            },
        )
        if not res_ajax.ok:
            raise PetitlyricsLyricPageParserError from ConnectionError
        bs_ajax = BeautifulSoup(res_ajax.text, "lxml")
        lyric_lines = list(PetitlyricsLyricData.model_validate_json(bs_ajax.text))

        return PetitlyricsLyricPage(
            title=select_one_tag(bs, "div.title-bar").text.strip(),
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=WithUrlText(text=artist.text.strip(), link=parse_obj_as_url(str(artist_link), base=url)),
            composer=None if composer is None else composer.strip(),
            lyricist=None if lyricist is None else lyricist.strip(),
            arranger=None,
            lyric_sections=convert_lines_into_sections(lyric_lines),
        )



"""<http://www.animap.jp>"""

from __future__ import annotations

import re

import requests
from bs4 import BeautifulSoup

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.util import get_captured_value, parse_obj_as_url, select_one_tag

from .model import AnimapLyricPage

_ANIMAP_PATTERN = r"^http://www\.animap\.jp/kasi/showkasi\.php\?surl=(?P<pageid>(?:ani|k-)\d+(?:-\d+)*)$"


class AnimapLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class AnimapLyricPageParser(BaseLyricPageParser):
    """http://www.animap.jp/kasi/showkasi.php?surl=<pageid>"""

    _test = "http://www.animap.jp/kasi/showkasi.php?surl=k-180131-061"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_ANIMAP_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> AnimapLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_ANIMAP_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise AnimapLyricPageParserError from ValueError

        bs = BeautifulSoup(requests.get(url, timeout=10).text.encode("shift_jis", "ignore"), "lxml")
        if bs is None:
            raise AnimapLyricPageParserError from ConnectionError

        lyric = select_one_tag(bs, "td.noprint.kasi_honbun")
        for br in lyric.select("br"):
            br.replace_with("\n")

        return AnimapLyricPage(
            title=select_one_tag(bs, "tr:nth-child(2) > td:nth-child(2)").text.strip(),
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=select_one_tag(bs, "tr:nth-child(1) > td:nth-child(2)").text.strip(),
            composer=select_one_tag(bs, "tr:nth-child(2) > td:nth-child(4)").text.strip(),
            lyricist=select_one_tag(bs, "tr:nth-child(1) > td:nth-child(4)").text.strip(),
            arranger=None,
            lyric_sections=[section.strip().split("\n") for section in lyric.text.strip().split("\n\n")],
        )



from __future__ import annotations

from base64 import standard_b64decode
from typing import TYPE_CHECKING, Annotated

from pydantic import BaseModel, BeforeValidator, RootModel, field_validator

from pyjlyric.model import LyricPage, WithUrlText

if TYPE_CHECKING:
    from collections.abc import Iterator


class PetitlyricsLyricPage(LyricPage):
    artist: WithUrlText
    composer: str | None = None
    lyricist: str | None = None
    arranger: None


class _PetitlyricsLyric(BaseModel):
    lyrics: str

    @field_validator("lyrics")
    @classmethod
    def validate_lyrics(cls, v: str) -> str:
        return standard_b64decode(v).decode("utf-8")


def _validate_root(v: list[dict[str, str]]) -> list[str]:
    return [_PetitlyricsLyric(**i).lyrics for i in v]


_PetitlyricsLyricDataDict = Annotated[list[str], BeforeValidator(_validate_root)]


class PetitlyricsLyricData(RootModel[_PetitlyricsLyricDataDict]):
    def __iter__(self) -> Iterator[str]:  # type: ignore[override]
        yield from self.root



"""<https://kashinavi.com>"""

from __future__ import annotations

import re

from bs4 import NavigableString

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.model import WithUrlText
from pyjlyric.util import get_captured_value, get_source, parse_obj_as_url, select_one_tag

from .model import KashinaviLyricPage

_KASHINAVI_PATTERN = r"^https://kashinavi\.com/song_view\.html\?(?P<pageid>\d+)"


class KashinaviLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class KashinaviLyricPageParser(BaseLyricPageParser):
    """https://kashinavi.com/song_view.html?<pageid>"""

    _test = "https://kashinavi.com/song_view.html?155779"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_KASHINAVI_PATTERN)
        m = re.match(pattern, url)
        return get_captured_value(m, "pageid") is not None

    @staticmethod
    def parse(url: str) -> KashinaviLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_KASHINAVI_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise KashinaviLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise KashinaviLyricPageParserError from ConnectionError

        artist_link = select_one_tag(bs, "body > center > p > a:nth-child(2)").get("href")
        if not isinstance(artist_link, str):
            raise KashinaviLyricPageParserError from ValueError

        h2_tag = select_one_tag(bs, "h2")
        title = h2_tag.text.strip()

        detail_tag = select_one_tag(bs, "div[style*='text-align:right'] > div[style*='text-align:left']")

        detail_text = detail_tag.text
        if not (
            m := re.match(
                r"^歌手：(.+?)作詞：(.+?)作曲：(.+)$",  # noqa: RUF001
                detail_text,
                flags=re.DOTALL,
            )
        ):
            raise KashinaviLyricPageParserError from ValueError
        artist, lyricist, composer = [i.strip() for i in m.groups()]

        return KashinaviLyricPage(
            title=title,
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=WithUrlText(link=parse_obj_as_url(artist_link), text=artist),
            composer=composer,
            lyricist=lyricist,
            arranger=None,
            lyric_sections=[
                [i.text for i in section.children if isinstance(i, NavigableString)]
                for section in select_one_tag(bs, "div[unselectable='on;']").find_all("p")
            ],
        )



"""<http://www.utamap.jp>"""

from __future__ import annotations

import re

from pyjlyric.base import BaseLyricPageParser, BaseLyricPageParserError
from pyjlyric.utamap.model import UtamapLyricPage
from pyjlyric.util import get_captured_value, get_source, parse_obj_as_url, select_one_tag

_UTAMAP_PATTERN = r"^https://www\.utamap\.com/showtop\.php\?surl=(?P<pageid>\d+)$"


class UtamapLyricPageParserError(BaseLyricPageParserError):
    """WIP."""


class UtamapLyricPageParser(BaseLyricPageParser):
    """http://www.utamap.com/showtop.php?surl=<pageid>"""

    _test = "https://www.utamap.com/showtop.php?surl=58401"

    @staticmethod
    def is_valid_url(url: str) -> bool:
        """Check if the url is valid."""
        pattern = re.compile(_UTAMAP_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")

        return pageid is not None

    @staticmethod
    def parse(url: str) -> UtamapLyricPage:
        """Parse the url page and return the result as LyricPage instance."""
        pattern = re.compile(_UTAMAP_PATTERN)
        m = re.match(pattern, url)
        pageid = get_captured_value(m, "pageid")
        if pageid is None:
            raise UtamapLyricPageParserError from ValueError

        bs = get_source(url)
        if bs is None:
            raise UtamapLyricPageParserError from ConnectionError

        lyric = select_one_tag(bs, "td.noprint.kasi_honbun")
        for br in lyric.select("br"):
            br.replace_with("\n")

        return UtamapLyricPage(
            title=select_one_tag(bs, "td.kasi1").text,
            page_url=parse_obj_as_url(url),
            pageid=pageid,
            artist=select_one_tag(bs, "tr:nth-child(3) > td:nth-child(2).pad5x10x0x10 > strong").text.strip(),
            composer=select_one_tag(bs, "tr:nth-child(2) > td:nth-child(2).pad5x10x0x10").text.strip(),
            lyricist=select_one_tag(bs, "tr:nth-child(1) > td:nth-child(2).pad5x10x0x10").text.strip(),
            arranger=None,
            lyric_sections=[section.strip().split("\n") for section in lyric.text.strip().split("\n\n")],
        )
